package ru.dbotthepony.kstarbound.network

import io.netty.buffer.ByteBuf
import io.netty.buffer.ByteBufInputStream
import io.netty.buffer.ByteBufOutputStream
import io.netty.channel.ChannelHandlerContext
import io.netty.channel.ChannelInboundHandlerAdapter
import io.netty.channel.ChannelOutboundHandlerAdapter
import io.netty.channel.ChannelPromise
import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import org.apache.logging.log4j.LogManager
import java.io.DataInputStream
import java.io.DataOutputStream
import kotlin.reflect.KClass

class PacketMapper {
	private val packets = ArrayList<Type<*>>()
	private val clazz2Type = Reference2ObjectOpenHashMap<KClass<*>, Type<*>>()

	private data class Type<T : IPacket>(val id: Int, val type: KClass<T>, val factory: IPacketReader<T>, val direction: PacketDirection)

	val size: Int
		get() = packets.size

	fun <T : IPacket> add(type: KClass<T>, reader: IPacketReader<T>, direction: PacketDirection = PacketDirection.get(type)): PacketMapper {
		if (packets.size >= 255)
			throw IndexOutOfBoundsException("Unable to add any more packet types! 255 is the max")

		if (type in clazz2Type)
			throw IllegalArgumentException("Already has packet handler for type $reader (${type})")

		val ptype = Type(packets.size, type, reader, direction)
		packets.add(ptype)
		clazz2Type[type] = ptype
		return this
	}

	inline fun <reified T : IPacket> add(reader: IPacketReader<T>, direction: PacketDirection = PacketDirection.get(T::class)): PacketMapper {
		return add(T::class, reader, direction)
	}

	inner class Inbound(val side: ConnectionSide) : ChannelInboundHandlerAdapter() {
		override fun channelRead(ctx: ChannelHandlerContext, msg: Any) {
			if (msg is ByteBuf) {
				val packetType = msg.readUnsignedByte().toInt()
				val type = packets.getOrNull(packetType)

				if (type == null) {
					LOGGER.error("Unknown packet type $packetType!")
					msg.release()
				} else if (!type.direction.acceptedOn(side)) {
					LOGGER.error("Packet ${type.type} can not be accepted on side $side!")
					msg.release()
				} else {
					try {
						ctx.fireChannelRead(type.factory.read(DataInputStream(ByteBufInputStream(msg))))
					} catch (err: Throwable) {
						LOGGER.error("Error while reading incoming packet from network", err)
					} finally {
						msg.release()
					}
				}
			} else {
				super.channelRead(ctx, msg)
			}
		}
	}

	inner class Outbound(val side: ConnectionSide) : ChannelOutboundHandlerAdapter() {
		override fun write(ctx: ChannelHandlerContext, msg: Any, promise: ChannelPromise) {
			val type = clazz2Type[msg::class]

			if (type == null) {
				LOGGER.error("Unknown outgoing message type ${msg::class}, it will not reach the other side.")
			} else if (!type.direction.acceptedOn(side.opposite)) {
				LOGGER.error("Packet ${type.type} can not be accepted on side ${side.opposite}, refusing to send it!")
			} else {
				val buff = ctx.alloc().buffer(2048)
				buff.writeByte(type.id)
				(msg as IPacket).write(DataOutputStream(ByteBufOutputStream(buff)))
				ctx.write(buff, promise)
			}
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
