package ru.dbotthepony.kstarbound.server.world

import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectAVLTreeSet
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.server.StarboundServer
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.WorldGeometry
import java.io.Closeable
import java.util.concurrent.atomic.AtomicInteger
import java.util.concurrent.locks.LockSupport
import kotlin.concurrent.withLock

class ServerWorld(
	val server: StarboundServer,
	seed: Long,
	geometry: WorldGeometry,
) : World<ServerWorld, ServerChunk>(seed, geometry), Closeable {
	init {
		server.worlds.add(this)
	}

	val thread = Thread(::runThread, "Starbound Server World $seed")
	var isStopped: Boolean = false
		private set

	init {
		thread.isDaemon = true
	}

	@Volatile
	private var nextThink = 0L

	override fun close() {
		if (!isStopped) {
			isStopped = true
			lock.withLock { onUnload() }
		}
	}

	fun startThread() {
		nextThink = System.nanoTime()
		thread.start()
	}

	private fun onUnload() {

	}

	private fun runThread() {
		while (!isStopped) {
			var diff = System.nanoTime() - nextThink

			while (diff < Starbound.TICK_TIME_ADVANCE_NANOS) {
				mailbox.executeQueuedTasks()
				diff = System.nanoTime() - nextThink
				LockSupport.parkNanos(diff)
				diff = System.nanoTime() - nextThink
			}

			nextThink = System.nanoTime()

			try {
				think()
			} catch (err: Throwable) {
				isStopped = true
				onUnload()
				throw err
			}
		}
	}

	override val isRemote: Boolean
		get() = false

	override fun thinkInner() {
		ticketLists.forEach { it.tick() }
	}

	override fun chunkFactory(pos: ChunkPos): ServerChunk {
		return ServerChunk(this, pos)
	}

	private val ticketMap = Long2ObjectOpenHashMap<TicketList>()
	private val ticketLists = ArrayList<TicketList>()

	interface ITicket {
		fun cancel()
		val isCanceled: Boolean
		val pos: ChunkPos
		val id: Int
	}

	interface ITimedTicket : ITicket, Comparable<ITimedTicket> {
		val timeRemaining: Int
		fun prolong(ticks: Int)

		override fun compareTo(other: ITimedTicket): Int {
			val cmp = timeRemaining.compareTo(other.timeRemaining)
			if (cmp != 0) return cmp
			return id.compareTo(other.id)
		}
	}

	private inner class TicketList(val pos: ChunkPos) {
		init {
			lock.withLock {
				check(ticketMap.put(pos.toLong(), this) == null) { "Already had ticket list at $pos" }
				ticketLists.add(this)
			}
		}

		private val permanent = ArrayList<Ticket>()
		private val temporary = ObjectAVLTreeSet<TimedTicket>()
		private var ticks = 0
		private var nextTicketID = AtomicInteger()

		fun tick(): Boolean {
			ticks++

			while (temporary.isNotEmpty() && temporary.first().timeRemaining <= 0) {
				val ticket = temporary.first()
				ticket.isCanceled = true
				temporary.remove(ticket)
			}

			return temporary.isNotEmpty() || permanent.isNotEmpty()
		}

		open inner class Ticket : ITicket {
			final override val id: Int = nextTicketID.getAndIncrement()
			final override val pos: ChunkPos
				get() = this@TicketList.pos

			final override fun cancel() {
				if (isCanceled) return

				lock.withLock {
					if (isCanceled) return
					isCanceled = true

					if (this is TimedTicket)
						temporary.remove(this)
					else
						permanent.remove(this)
				}
			}

			final override var isCanceled: Boolean = false
		}

		inner class TimedTicket(var expiresAt: Int) : Ticket(), ITimedTicket {
			override val timeRemaining: Int
				get() = (expiresAt - ticks).coerceAtLeast(0)

			override fun prolong(ticks: Int) {
				if (ticks == 0 || isCanceled) return

				lock.withLock {
					if (isCanceled) return

					temporary.remove(this)
					expiresAt += ticks
					if (timeRemaining > 0) temporary.add(this)
				}
			}
		}
	}
}
